; Hardware/driver routines for BMOW's built-in LCD display

; memory-mapped addresses
LCD_ADR_DATA = $003F00
LCD_ADR_CONTROL = $003F01

; dimensional constants
LCD_WIDTH = 24
LCD_HEIGHT = 2

; driver RAM addresses
LCD_ADR_DISPLAY_BASE = DRIVER_RAM_BASE
LCD_ADR_DISPLAY_LINE_1 = LCD_ADR_DISPLAY_BASE
LCD_ADR_DISPLAY_LINE_2 = LCD_ADR_DISPLAY_BASE + (LCD_WIDTH * 1)
LCD_ADR_CURSOR = LCD_ADR_DISPLAY_BASE + (LCD_WIDTH * LCD_HEIGHT)
LCD_WRITE_DECIMAL_IN_L = LCD_ADR_CURSOR+1
LCD_WRITE_DECIMAL_IN_H = LCD_ADR_CURSOR+2
LCD_WRITE_DECIMAL_IN_B = LCD_ADR_CURSOR+3
LCD_WRITE_DECIMAL_OUT_L = LCD_ADR_CURSOR+4
LCD_WRITE_DECIMAL_OUT_H = LCD_ADR_CURSOR+5
LCD_WRITE_DECIMAL_OUT_B = LCD_ADR_CURSOR+6
LCD_WRITE_DECIMAL_OUT_BB = LCD_ADR_CURSOR+7
LCD_WRITE_DECIMAL_FLAG = LCD_ADR_CURSOR+8

LCD_ADR_END = LCD_ADR_CURSOR

; command constants
LCD_CLEAR_DISPLAY_COMMAND = $01

LCD_RETURN_HOME_COMMAND = $02

LCD_DISPLAY_COMMAND = $08
LCD_DISPLAY_ON = $04
LCD_DISPLAY_OFF = $00
LCD_CURSOR_ON = $02
LCD_CURSOR_OFF = $00
LCD_CURSOR_BLINK_ON = $01
LCD_CURSOR_BLINK_OFF = $00

LCD_FUNCTION_SET_COMMAND = $20
LCD_FUNCTION_SET_8BIT = $10
LCD_FUNCTION_SET_4BIT = $00
LCD_FUNCTION_SET_2LINE = $08
LCD_FUNCTION_SET_1LINE = $00
LCD_FUNCTION_SET_5x10FONT = $04
LCD_FUNCTION_SET_5x8FONT = $00

LCD_CGRAM_ADDRESS_COMMAND = $40
LCD_CGRAM_ADDRESS_MASK = $3F

LCD_DISPLAY_ADDRESS_COMMAND = $80
LCD_DISPLAY_ADDRESS_MASK = $7F

; timing constants
LCD_CLEAR_HOME_DELAY_US = 1520
LCD_SINGLE_COMMAND_DELAY_US = 37
LCD_CLOCKS_PER_US = 2

; --- LCD delay single
; A macro to busy-wait after a single LCD command.
!macro lcd_delay_single {	
	; the delay loop takes 5 clock cycles
	lda #((LCD_SINGLE_COMMAND_DELAY_US * LCD_CLOCKS_PER_US) / 5)
-	dec
	bne -
}


; --- LCD delay clear/home
; A macro to busy-wait after a clear/home LCD command.
!macro lcd_delay_clear_home {	
	; the delay loop takes 5 clock cycles
	!for Outer, ((((LCD_CLEAR_HOME_DELAY_US * LCD_CLOCKS_PER_US) / 5) / 255) + 1) {
		lda #$FF
-		dec
		bne -	
	}
}


; --- LCD command
; @param0 - A one-byte command value formed by logical-ORing a single LCD command and related state constants.
; Do not use lcd_command for clear or home of the LCD: the delay won't be correct.
lcd_command:
	sta LCD_ADR_CONTROL
	+lcd_delay_single
	rts
	
	
; --- LCD clear
; @noparams
; Clears the entire LCD display, and sets the cursor position to the top-left corner.
lcd_clear:
	lda #LCD_CLEAR_DISPLAY_COMMAND
	sta LCD_ADR_CONTROL
	lda #0
	sta LCD_ADR_CURSOR
	+lcd_delay_clear_home
	
	; set the display memory to all spaces
	lda #32
	ldx #(LCD_WIDTH*LCD_HEIGHT-1)
-	sta LCD_ADR_DISPLAY_BASE,X
	dex
	bpl -
	
	rts	


; --- LCD home
; @noparams
; Clears the entire LCD display, and sets the cursor position to the top-left corner.
lcd_home:
	lda #LCD_RETURN_HOME_COMMAND
	sta LCD_ADR_CONTROL
	lda #0
	sta LCD_ADR_CURSOR	
	+lcd_delay_clear_home
	rts	
	
	
; --- LCD standard init
; @noparams
; Clears the LCD display, activates the cursor, and selects 2-line display mode.
lcd_standard_init:
	lda #(LCD_FUNCTION_SET_COMMAND | LCD_FUNCTION_SET_8BIT | LCD_FUNCTION_SET_2LINE | LCD_FUNCTION_SET_5x8FONT)
	jsr lcd_command
	lda #(LCD_DISPLAY_COMMAND | LCD_DISPLAY_ON | LCD_CURSOR_OFF)
	jsr lcd_command
	jsr lcd_clear
	lda #0
	sta LCD_ADR_CURSOR	
	rts
		

; --- LCD newline
; @noparams
; Sets the cursor to the beginning of the next line. If the cursor is currently on the
; last line, the display will be scrolled.
lcd_write_newline:
	ldx LCD_ADR_CURSOR
	cpx #24
	bpl +
	; set new logical cursor position to 24
	ldx #24
	bra ++
	
+	; scroll the display
	jsr lcd_scroll
	; set new logical cursor position to 24
	ldx #24

++	stx LCD_ADR_CURSOR
	jsr lcd_sync_cursor
	rts
	
	
; --- LCD backspace
; @noparams
; Moves the cursor backwards by one logical position. Prints a space at the new position.
lcd_backspace:
	; back up the cursor
	ldx LCD_ADR_CURSOR
	dex
	bmi +
	stx LCD_ADR_CURSOR
	jsr lcd_sync_cursor

	; write a space	
+	lda #32
	jsr lcd_write_char
	
	; back up the cursor again
	ldx LCD_ADR_CURSOR
	dex
	stx LCD_ADR_CURSOR
	jsr lcd_sync_cursor
	rts
	
		
; --- LCD write character
; @param0 - A one byte ASCII character value, to be displayed at the current cursor position.
; The display will be scrolled if necessary. Return is handled as well.	
lcd_write_char:	
	; return character?
	cmp #13
	bne +
	jsr lcd_write_newline
	rts
	
	; backspace?
+	cmp #8
	bne +
	jsr lcd_backspace
	rts
		
+	sta LCD_ADR_DATA
	ldx LCD_ADR_CURSOR
	sta LCD_ADR_DISPLAY_BASE,X
	inx
	stx LCD_ADR_CURSOR
	+lcd_delay_single
	
	cpx #24
	bne +
	; at logical position 24, jump to physical position 64
	lda #(LCD_DISPLAY_ADDRESS_COMMAND | 64) 
	bra ++

+	cpx #48
	bne +++
	; at logical position 48, scroll the display by one line
	jsr lcd_scroll
	bra +++
					
++	sta LCD_ADR_CONTROL
	+lcd_delay_single
	
+++	rts


; --- LCD scroll
; @noparams
; Scrolls the contents of the LCD up by one line, and sets the cursor to the beginning
; of the last line.
lcd_scroll:
	; copy line 2 of the display buffer to line 1
	ldx #0
-	lda LCD_ADR_DISPLAY_LINE_2,X
	sta LCD_ADR_DISPLAY_LINE_1,X
	inx
	cpx #(LCD_WIDTH * (LCD_HEIGHT - 1))
	bne -
	
	; clear the last line
	ldx #(LCD_WIDTH - 1)
	lda #' '
-	sta	LCD_ADR_DISPLAY_LINE_2,X
	dex
	bpl -
	beq - ; TODO: remove after microcode is fixed
	
	; set cursor to beginning of the last line
	lda #(LCD_WIDTH * (LCD_HEIGHT - 1))
	sta LCD_ADR_CURSOR

	; refresh the LCD from the display buffer state
	jsr lcd_refresh
	rts
	
			
; --- LCD sync hardware cursor
; @noparams
; Synchronizes the LCD's hardware cursor with LCD_ADR_CURSOR.
lcd_sync_cursor:
	lda LCD_ADR_CURSOR
	cmp #24
	; if logical position is less than 24, use it as the physical position
	bmi ++
	
+	; if logical position is 24-48, add 40 to get the physical position
	clc
	adc #40
	
++	ora #LCD_DISPLAY_ADDRESS_COMMAND
	sta LCD_ADR_CONTROL
	+lcd_delay_single
	rts


; --- LCD refresh
; @noparams
; Refreshes the LCD from the display buffer state.
lcd_refresh:
	; set the cursor to the beginning of the first line
	lda #(LCD_DISPLAY_ADDRESS_COMMAND | 0)
	sta LCD_ADR_CONTROL
	+lcd_delay_single
	
	; copy the first line
	ldx #0
-	lda LCD_ADR_DISPLAY_LINE_1,X
	sta LCD_ADR_DATA
	+lcd_delay_single
	inx
	cpx #LCD_WIDTH
	bne -
		
	; set the cursor to the beginning of the second line
	lda #(LCD_DISPLAY_ADDRESS_COMMAND | 64)
	sta LCD_ADR_CONTROL
	+lcd_delay_single
		
	; copy the second line
	ldx #0
-	lda LCD_ADR_DISPLAY_LINE_2,X
	sta LCD_ADR_DATA
	+lcd_delay_single
	inx
	cpx #LCD_WIDTH
	bne -
				
	; set the hardware cursor position
	jsr lcd_sync_cursor
	rts
	
	
; --- LCD write hex nibble
; @param0 - A value less than 16, to be displayed as a signle hex digit at the current cursor position.
; The display will be scrolled if necessary.	
lcd_write_hex_nibble:
	cmp #10
	bmi +
	
	; A-F
	clc
	adc #55
	jsr lcd_write_char
	bra ++
		
	; 0-9
+	clc
	adc #'0'
	jsr lcd_write_char
	
++  rts

	
; --- LCD write hex byte
; @param0 - A one byte value, to be displayed as two hex digits, beginning at the current cursor position.
; The display will be scrolled if necessary.	
lcd_write_hex:	
	; store A
	pha
	
	; A = A >> 4
	; ROL * 5 followed by AND should be faster than LSR * 4
	rol
	rol
	rol
	rol
	rol
	and #$F
	jsr lcd_write_hex_nibble
	
	; restore A, A = A & 0xF
	pla
	and #$F
	jsr lcd_write_hex_nibble
	
	rts

	